# -*- coding: utf-8 -*-

# Standard Library
import json
import datetime
from collections import OrderedDict

# Project Library
from kinder.core.unit.models import FilialData, FilialTypeStatus
from gisdo import constants as const
from gisdo.exceptions import NoDataInMo
from gisdo.alchemy_session import session_scope
from gisdo.utils import DateMixin
from gisdo.models import ReportFormRow
from gisdo.index import (
    ApplicationIndex,
    EnrolledIndex,
    DirectIndex,
    CapacitiesIndex,
    get_queue_index,
    get_enrolled_age_filter,
    get_enrolled_count,
    get_queue_age_filter,
    get_queue_count,
    get_queue_index_collection)


class ReportCacheMixin(object):

    APPLICATIONS = 'applications'
    QUEUES = 'queues'
    ENROLLED = 'enrolled'
    CAPACITIES = 'capacities'

    DATA_TYPE = [
        APPLICATIONS,
        QUEUES,
        ENROLLED,
        CAPACITIES
    ]

    def __init__(self):

        self._cache = {
            self.APPLICATIONS: {},
            self.QUEUES: {},
            self.ENROLLED: {},
            self.CAPACITIES: {}
        }

    def _update_cache(self, data, data_type):
        u"""Обновление кэша."""

        for key, value in data.iteritems():
            if not key in self._cache[data_type]:
                self._cache[data_type][key] = value
            elif isinstance(value, dict):
                for age_cat in data[key]:
                    self._cache[
                        data_type][key][age_cat] += data[key][age_cat]


class DataCollector(ReportCacheMixin):
    def __init__(self, mo, unit_helper):

        super(DataCollector, self).__init__()

        self._mo = mo
        self._unit_helper = unit_helper

    def collect(self, report_form):

        # Признак наличия садов по которым можно
        # собрать данные
        dou_list = self._unit_helper.get_mo_units(self._mo.id)
        have_data = len(dou_list)

        if not have_data:
            raise NoDataInMo()

        # Собираем по ДОУ
        for dou in dou_list:
            self._dou_collect(report_form, dou)

        # Данные по МО лежат в кэше
        ReportFormRow.objects.get_or_create(
            applications=json.dumps(self._cache[self.APPLICATIONS]),
            queues=json.dumps(self._cache[self.QUEUES]),
            enrolled=json.dumps(self._cache[self.ENROLLED]),
            capacities=json.dumps(self._cache[self.CAPACITIES]),
            report=report_form,
            unit=self._mo)

        return self._cache

    def _dou_collect(self, report_form, dou):

        applications = self._collect_applications_data(dou)
        queues = self._collect_queue_data(dou)
        enrolled = self._collect_enrolled_data(dou)
        capacities = self._collect_capacities(dou)

        ReportFormRow.objects.get_or_create(
            report=report_form, unit=dou,
            applications=json.dumps(applications),
            queues=json.dumps(queues),
            enrolled=json.dumps(enrolled),
            capacities=json.dumps(capacities))

    @staticmethod
    def _merge_dicts(first_dict, second_dict):
        for index in first_dict:
            for age_cat in first_dict[index]:
                try:
                    value = second_dict[index][age_cat]
                except KeyError:
                    value = 0
                first_dict[index][age_cat] += value

    def _collect_applications_data(self, dou):

        def _inner(dou):

            # Словарь с id-шниками показателей
            applications = OrderedDict([
                ('1', OrderedDict()),
                ('1.1', OrderedDict()),
                ('2', OrderedDict()),
                ('3', OrderedDict()),
                ('4', OrderedDict()),
                ('4.1', OrderedDict()),
                ('4.2', OrderedDict()),
                ('5', OrderedDict()),
                ('6', OrderedDict()),
            ])

            index_4 = get_queue_index_collection("4", dou)
            index_4_1 = get_queue_index_collection("4.1", dou)
            index_4_2 = get_queue_index_collection("4.2", dou)
            index_5 = get_queue_index_collection("5", dou)

            with session_scope() as session:
                index = ApplicationIndex(dou, session)

                for value in const.AGE_CATEGORIES_FULL.values():

                    applications['2'][value] = index.get_count(
                        index_type=ApplicationIndex.COUNT_BY_DELIVERY,
                        portal=1,
                        age_range=value, on_date=datetime.date.today)
                    applications['3'][value] = index.get_count(
                        index_type=ApplicationIndex.COUNT_BY_DELIVERY,
                        age_range=value, on_date=datetime.date.today)
                    applications['1.1'][value] = (
                        applications['2'][value] + applications['3'][value])

                    on_date_filter = get_queue_age_filter(
                        value, datetime.date.today)
                    applications["4"][value] = get_queue_count(
                        index_4, on_date_filter)
                    applications["4.1"][value] = get_queue_count(
                        index_4_1, on_date_filter)
                    applications["4.2"][value] = get_queue_count(
                        index_4_2, on_date_filter)
                    applications["5"][value] = get_queue_count(
                        index_5, on_date_filter)

                    # Общее количество заявлений на постановку в очередь
                    # в группы кратковременного пребывания
                    applications['6'][value] = index.get_count(
                        index_type=ApplicationIndex.COUNT_BY_GROUP_TYPE,
                        age_range=value, on_date=datetime.date.today)

                applications['1'][const.ALL] = get_queue_index('1', dou)

            return applications

        app = _inner(dou)
        for filial in self._unit_helper.get_filial(dou):
            filial_app = _inner(filial)
            self._merge_dicts(app, filial_app)

        self._update_cache(app, self.APPLICATIONS)

        return app

    def _collect_queue_data(self, dou):

        def _inner(dou):
            queues = OrderedDict([
                ('7', OrderedDict()), ('7.1', OrderedDict()),
                ('7.2', OrderedDict()), ('7.3', OrderedDict()),
                ('7.4', OrderedDict()), ('8', OrderedDict()),
                ('9', OrderedDict()), ('9.1', OrderedDict()),
                ('10', OrderedDict()), ('11', OrderedDict()),
                ('12', OrderedDict()), ('12.1', OrderedDict()),
                ('13', OrderedDict()),
                ('13.1', OrderedDict()), ('14', OrderedDict()),
                ('15', OrderedDict()), ('16', OrderedDict()),
                ('17', OrderedDict()), ('17.1', OrderedDict()),
                ('18', OrderedDict()), ('18.1', OrderedDict()),
                ('18.2', OrderedDict()), ('18.3', OrderedDict())
            ])

            index_7 = get_queue_index_collection("7", dou)
            index_7_1 = get_queue_index_collection("7.1", dou)
            index_7_2 = get_queue_index_collection("7.2", dou)
            index_7_3 = get_queue_index_collection("7.3", dou)
            index_7_4 = get_queue_index_collection('7.4', dou)
            index_18_1 = get_queue_index_collection('18.1', dou)
            index_18_3 = get_queue_index_collection('18.3', dou)
            index_20 = get_queue_index_collection("20", dou)
            index_12 = get_queue_index_collection("12", dou)
            index_12_1 = get_queue_index_collection("12.1", dou)
            index_13 = get_queue_index_collection("13", dou)
            index_13_1 = get_queue_index_collection("13.1", dou)
            index_14 = get_queue_index_collection("14", dou)
            index_15 = get_queue_index_collection("15", dou)
            index_16 = get_queue_index_collection("16", dou)
            index_17 = get_queue_index_collection("17", dou)
            index_17_1 = get_queue_index_collection("17.1", dou)
            index_18 = get_queue_index_collection("18", dou)
            index_18_2 = get_queue_index_collection("18.2", dou)
            index_10 = get_queue_index_collection("10", dou)
            index_11 = get_queue_index_collection("11", dou)

            queues['10'][const.ALL] = 0
            queues['11'][const.ALL] = 0
            for value in const.AGE_CATEGORIES_FULL.values():

                on_date_filter = get_queue_age_filter(
                    value, datetime.date.today)
                current_year_first_september = get_queue_age_filter(
                    value,
                    DateMixin.get_current_calendar_year())
                next_year_first_day = get_queue_age_filter(
                    value,
                    datetime.date(datetime.date.today().year+1, 1, 1))

                # Отложенный спрос
                queues["7"][value] = get_queue_count(
                    index_7, current_year_first_september)
                queues["7.1"][value] = get_queue_count(
                    index_7_1, on_date_filter)
                queues["7.2"][value] = get_queue_count(
                    index_7_2, on_date_filter)
                queues["7.4"][value] = get_queue_count(
                    index_7_4, on_date_filter)
                # Актуальный спрос
                queues["18.1"][value] = get_queue_count(
                    index_18_1, on_date_filter)
                queues["18.3"][value] = get_queue_count(
                    index_18_3, on_date_filter)
                queues["7.3"][value] = get_queue_count(
                    index_7_3, on_date_filter)

                # Актуальный спрос + Желает сменить ДОО
                index20 = get_queue_count(index_20, on_date_filter)
                queues['8'][value] = (
                    queues['18.1'][value] + queues['18.3'][value] +
                    index20)

                # Количество детей в очереди с ограниченными
                # возможностями по здоровью
                queues["12"][value] = get_queue_count(index_12, on_date_filter)
                queues["12.1"][value] = get_queue_count(
                    index_12_1, on_date_filter)
                queues["13"][value] = get_queue_count(index_13, on_date_filter)
                queues["13.1"][value] = get_queue_count(
                    index_13_1, on_date_filter)
                queues["14"][value] = get_queue_count(
                    index_14, on_date_filter)
                queues["15"][value] = get_queue_count(index_15, on_date_filter)
                queues["16"][value] = get_queue_count(index_16, on_date_filter)
                queues["17"][value] = get_queue_count(index_17, on_date_filter)
                queues["17.1"][value] = get_queue_count(
                    index_17_1, on_date_filter)

                # Актуальный спрос
                queues["18"][value] = get_queue_count(
                    index_18, current_year_first_september)
                queues["18.2"][value] = get_queue_count(
                    index_18_2, next_year_first_day)

                queues["9"][value] = get_queue_count(
                    index_10, current_year_first_september
                ) + get_queue_count(index_11, current_year_first_september)

                # Количество детей в очереди имеющих право зачисления
                # на основе федеральных и региональных льгот.
                current_10_value = get_queue_count(index_10, on_date_filter)
                queues['10'][const.ALL] += current_10_value
                current_11_value = get_queue_count(index_11, on_date_filter)
                queues['11'][const.ALL] += current_11_value
                queues['9.1'][value] = (
                    current_10_value + current_11_value)
            return queues

        queues = _inner(dou)
        for filial in self._unit_helper.get_filial(dou):
            filial_queues = _inner(filial)
            self._merge_dicts(queues, filial_queues)

        self._update_cache(queues, self.QUEUES)

        return queues

    def _collect_enrolled_data(self, dou):

        def _inner(dou):
            enrolled = OrderedDict([
                ('19', OrderedDict()), ('19.1', OrderedDict()),
                ('19.2', OrderedDict()), ('19.3', OrderedDict()),
                ('20', OrderedDict()), ('20.1', OrderedDict()),
                ('20.2', OrderedDict()), ('20.3', OrderedDict()),
                ('20.4', OrderedDict()), ('21', OrderedDict()),
                ('22', OrderedDict()), ('23', OrderedDict()),
                ('24', OrderedDict()), ('25', OrderedDict()),
                ('26', OrderedDict()), ('27', OrderedDict()),
                ('28', OrderedDict()), ('29', OrderedDict()),
                ('30', OrderedDict()), ('30.1', OrderedDict()),
                ('30.2', OrderedDict()),
            ])

            with session_scope() as session:
                index = EnrolledIndex(dou, session)
                direct_index = DirectIndex(dou)

                index_18 = get_queue_index_collection("18", dou)
                index_18_1 = get_queue_index_collection('18.1', dou)
                index_18_2 = get_queue_index_collection("18.2", dou)
                index_19 = index.get_count()
                index_19_1 = index.get_count()
                index_19_2 = index.get_count()
                index_20 = get_queue_index_collection("20", dou)
                index_20_1 = get_queue_index_collection("20.1", dou)
                index_20_2 = get_queue_index_collection("20.2", dou)
                index_20_3 = get_queue_index_collection("20.3", dou)
                index_22 = index.get_count(
                    index_type=EnrolledIndex.COUNT_DISABLED)
                index_23 = index.get_count(
                    index_type=EnrolledIndex.COUNT_CORRECT)
                index_24 = index.get_count(
                    index_type=EnrolledIndex.COUNT_HEALTH)
                index_25 = index.get_count(
                    index_type=EnrolledIndex.COUNT_BY_CARE)
                index_26 = index.get_count(
                    index_type=EnrolledIndex.COUNT_FULL_DAY_STAY)
                index_27 = index.get_count(
                    index_type=EnrolledIndex.COUNT_SHORT_STAY)

                for value in const.AGE_CATEGORIES_FULL.values():
                    current_year_enrolled_filter = get_enrolled_age_filter(
                        value, DateMixin.get_current_calendar_year())
                    on_date_enrolled_filter = get_enrolled_age_filter(
                        value, datetime.date.today)
                    next_calendar_year_enrolled = get_enrolled_age_filter(
                        value, DateMixin.get_next_calendar_year)

                    # Общая численность детей зачисленных в ДОО
                    enrolled["19"][value] = get_enrolled_count(
                        index_19, current_year_enrolled_filter)
                    enrolled["19.1"][value] = get_enrolled_count(
                        index_19_1, on_date_enrolled_filter)
                    enrolled["19.2"][value] = get_enrolled_count(
                        index_19_2, next_calendar_year_enrolled)
                    enrolled['19.3'][value] = direct_index.get_count(
                        DirectIndex.FILTER_193, age_range=value,
                        on_date=datetime.date.today)

                    on_date_filter = get_queue_age_filter(
                        value, datetime.date.today)
                    enrolled["20"][value] = get_queue_count(
                        index_20, on_date_filter)
                    enrolled["20.1"][value] = get_queue_count(
                        index_20_1, on_date_filter)
                    enrolled["20.2"][value] = get_queue_count(
                        index_20_2, on_date_filter)
                    enrolled["20.3"][value] = get_queue_count(
                        index_20_3, on_date_filter)

                    # Общая численность детей, зачисленных в группы для
                    # детей с ограниченными возможностями здоровью
                    enrolled["22"][value] = get_enrolled_count(
                        index_22, on_date_enrolled_filter)

                    # Общее количество детей зачисленных в
                    # оздоровительные и компенсирующие группы
                    enrolled["23"][value] = get_enrolled_count(
                        index_23, on_date_enrolled_filter)
                    enrolled["24"][value] = get_enrolled_count(
                        index_24, on_date_enrolled_filter)
                    enrolled["25"][value] = get_enrolled_count(
                        index_25, on_date_enrolled_filter)

                    # Общая численность детей, зачисленных в
                    # группы круглосуточного пребывания детей
                    enrolled["26"][value] = get_enrolled_count(
                        index_26, on_date_enrolled_filter)

                    # Группы кратковременного пребывания
                    enrolled["27"][value] = get_enrolled_count(
                        index_27, on_date_enrolled_filter)

                    enrolled['28'][value] = enrolled['27'][value]

                    current_year_first_september = get_queue_age_filter(
                        value,
                        DateMixin.get_current_calendar_year())
                    enrolled['30'][value] = (
                        enrolled["19"][value] +
                        get_queue_count(index_18, current_year_first_september) +
                        direct_index.get_count(
                            DirectIndex.FILTER_193, age_range=value,
                            on_date=DateMixin.get_current_calendar_year
                        )
                    )

                    on_date_filter = get_queue_age_filter(
                        value, datetime.date.today)
                    enrolled['30.1'][value] = (
                        get_queue_count(index_18_1, on_date_filter) +
                        enrolled['19.1'][value] + enrolled['19.3'][value]
                    )

                    next_year_first_day = get_queue_age_filter(
                        value,
                        datetime.date(datetime.date.today().year+1, 1, 1))
                    enrolled['30.2'][value] = (
                        enrolled['19.2'][value] +
                        get_queue_count(index_18_2, next_year_first_day) +
                        direct_index.get_count(
                            DirectIndex.FILTER_193, age_range=value,
                            on_date=DateMixin.get_next_calendar_year
                        )
                    )

                enrolled['20.4'][const.ALL] = get_queue_index("20.4", dou)
                # Общее количество детей зачисленных в семейные,
                # оздоровительные и компенсирующие группы
                enrolled['21'][const.ALL] = index.get_count(
                    index_type=EnrolledIndex.COUNT_FAMILY,
                    on_date=datetime.date.today).count()
                enrolled['29'][const.ALL] = index.get_count(
                    index_type=EnrolledIndex.COUNT_DECREASE_CONTINGENT,
                    on_date=datetime.date.today).count()

            return enrolled

        enrolled = _inner(dou)
        for filial in self._unit_helper.get_filial(dou):
            filial_enrolled = _inner(filial)
            self._merge_dicts(enrolled, filial_enrolled)

        self._update_cache(enrolled, self.ENROLLED)

        return enrolled

    def _collect_capacities(self, dou):

        def _inner(dou):
            capacities = OrderedDict([
                ('31', OrderedDict()), ('31.1', OrderedDict()),
                ('31.2', OrderedDict()), ('31.3', OrderedDict()),
                ('31.4', OrderedDict()), ('32', OrderedDict()),
                ('33', OrderedDict())
            ])

            index = CapacitiesIndex(dou)
            direct_index = DirectIndex(dou)

            for value in const.AGE_CATEGORIES_EIGHT.values():
                capacities['31'][value] = index.get_count(
                    age_category=value)
                capacities['31.1'][value] = index.get_count(
                    age_category=value,
                    index_type=CapacitiesIndex.HEALTH_NEED)
                capacities['31.2'][value] = index.get_count(
                    age_category=value,
                    index_type=CapacitiesIndex.COMPENSATING)
                capacities['31.3'][value] = index.get_count(
                    age_category=value,
                    index_type=CapacitiesIndex.HEALTH)
                capacities['31.4'][value] = index.get_count(
                    age_category=value,
                    index_type=CapacitiesIndex.SHORT_STAY)

            capacities['32'][const.ALL] = direct_index.get_count(
                DirectIndex.FILTER_32,
                on_date=datetime.date.today)
            capacities['33'][const.ALL] = get_queue_index("33", dou)

            return capacities

        capacities = _inner(dou)
        for filial in self._unit_helper.get_filial(dou):
            filial_capacities = _inner(filial)
            self._merge_dicts(capacities, filial_capacities)

        self._update_cache(capacities, self.CAPACITIES)

        return capacities
